/*
 This file is part of the Simple Java Message Exchange (SIJAME) software.

(c) 2011 INSERM
SIJAME is developed at Montpellier RIO Imaging (IFR 122), Montpellier, France (www.mri.cnrs.fr)
Developer: Volker Baecker (volker.baecker@mri.cnrs.fr) 

This software is a very simple message oriented middleware (mom). It allows 
a communication between programms via sockets. A software (client) can send a 
message to another software on another machine (server). The message is 
put into a message queue. Whenever the currently first message in the queue changes
a notification is sent to observers of the queue.  

This software is governed by the CeCILL-B license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the CeCILL-B
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-B license and that you accept its terms. 
 */
package fr.cnrs.mri.sijame.connection;

import java.io.IOException;
import java.net.ServerSocket;
import java.net.Socket;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Observable;
import java.util.Set;
import java.util.logging.Logger;
import fr.cnrs.mri.sijame.messages.Message;
import fr.cnrs.mri.sijame.messages.MessageQueue;
import fr.cnrs.mri.util.logging.LoggingUtil;

/**
 * The server listens for client connections on a given port. It runs in a separate thread so
 * that is can be accessed while it listens on a socket. Incoming connections are handled in another thread so
 * that multiple connections can be handled at the same time.
 * 
 * @author baecker
 *
 */
public class SIJAMEServer extends Observable implements Runnable, MessageSource {
	public enum Aspect {IS_RUNNING}; 
	/**
	 * The port on which the server is listening.
	 */
	private int port;
	/**
	 * The server socket on which the server is contacted by clients.
	 */
	private ServerSocket serverSocket;
	/**
	 * A set of communication processors. A communication processor contains the socket and streams for a communication
	 * with a client. A communication processor is added when a client connects the server and removed when the communication
	 * is finished. 
	 */
	private Set<CommunicationProcessor> communications = new HashSet<CommunicationProcessor>();
	/**
	 * A flag indicating that the server has been shutdown by purpose. 
	 */
	private boolean shutdown;
	/**
	 * A flag indicating if the server is currently running, i.a. if it is currently listening for incoming connections.
	 */
	private boolean isRunning;
	/**
	 * The logger used for information and error output by the server.
	 */
	private Logger logger;
	private MessageQueue requestQueue;
	private MessageQueue answerQueue;
	private boolean aboutToShutdown;
	
	public static Map<Integer, SIJAMEServer> server;

	private List<MessageHandler> messageHandlers = new ArrayList<MessageHandler>();
	/**
	 * Create a new server that will use the given port to listen for connections once it has been started.
	 * 
	 * @param port	The port on which the server will accept connections.
	 */
	public SIJAMEServer(int port) {
		this.port = port;
		this.logger = LoggingUtil.getLoggerFor(this);
		getServer().put(port, this);
	}

	/**
	 * java SIJAMEServer.class <port>
	 * 
	 * @param args	the program expects the port as argument
	 */
	public static void main(String[] args) {
		if (args.length != 1) {
			System.err.println("Usage: java SIJAMEServer.class port_number]");
			System.exit(1);
		}
		int port = Integer.parseInt(args[0]);
		SIJAMEServer server = SIJAMEServer.currentForPort(port);
		server.start();
	}
	
	/**
	 * Start the server in a new thread, so that the object can answer messages while the server is listening for 
	 * connections and handling communications.
	 */
	public void start() {
		Thread thread = new Thread(this);
		thread.start();
	}
	
	/**
	 * Listen for connections and start a communication in a new thread when a client connects.
	 */
	@Override
	public void run() {
		shutdown = false;
		aboutToShutdown = false;
		this.setRunning(false);
		if (!this.createServerSocket()) return;
		this.setRunning(true);
		logger.info("Listening on port " + port + "...");
		while (!shutdown) {
			Socket socket;
			try {
				socket = serverSocket.accept();
				logger.info("Got connection from " + socket);
				CommunicationProcessor cp = new CommunicationProcessor(this, socket);
				communications.add(cp);
			} catch (IOException e) {
			  if (!aboutToShutdown) logger.warning(LoggingUtil.getMessageAndStackTrace(e));
			}
		}
		this.setRunning(false);
	}

	/**
	 * The port on which the server is accepting connections.
	 * 
	 * @return	The port on which the server is accepting connections.
	 */
	public int getPort() {
		return port;
	}
	
	public void shutdown() {
		if (!this.isRunning()) return; 
		this.aboutToShutdown  = true;
		this.getRequestQueue().deleteObservers();
		this.getAnswerQueue().deleteObservers();
		
		for (CommunicationProcessor communication : communications) 
			communication.closeStreams();
		try {
			if (serverSocket!=null) serverSocket.close();
		} catch (IOException e) {
			logger.warning(LoggingUtil.getMessageAndStackTrace(e));
		} 
		this.setRunning(false);
		shutdown = true;
		logger.info("Shutting down the server for port " + port);
	}
	
	/**
	 * Answer if the server is currently running, i.e. if he is currently listening for connections.
	 * 
	 * @return	True if the server is running and false if it has been stopped.
	 */
	public boolean isRunning() {
		return isRunning;
	}
	
	public MessageQueue getRequestQueue() {
		if (this.requestQueue==null) this.requestQueue = new MessageQueue();
		return requestQueue;
	}
	
	public MessageQueue getAnswerQueue() {
		if (this.answerQueue==null) this.answerQueue = new MessageQueue();
		return answerQueue;
	}
	
	public static SIJAMEServer currentForPort(int port) {
		if (!getServer().containsKey(port)) {
			new SIJAMEServer(port);
		}
		SIJAMEServer result = getServer().get(port); 
		return result;
	}
	
	/**
	 * When a communication finishes the communication processor 
	 * 
	 * @param communicationProcessor
	 */
	protected void removeConnection(CommunicationProcessor communicationProcessor) {
		communications.remove(communicationProcessor);
	}
	
	private void changed(Aspect anAspect) {
		this.setChanged();
		this.notifyObservers(anAspect);
		this.clearChanged();
	}
	
	/**
	 * Create a new server socket for the port on which the server will accept connections.
	 *
	 * @return	true if the socket has been created and false if an error occurred.
	 */
	private boolean createServerSocket() {
		boolean result = true;
		try {
			serverSocket = new ServerSocket(port);
		} catch (IOException e) {
			result = false;
			logger.warning(LoggingUtil.getMessageAndStackTrace(e));
		}
		return result;
	}
	
	synchronized static public Map<Integer, SIJAMEServer> getServer() {
		if (server==null) {
			server = new HashMap<Integer, SIJAMEServer>();
		}
		return server;
	}
	
	@Override
	public void addMessageHandler(MessageHandler handler) {
		if (this.messageHandlers.contains(handler)) return;
		handler.setMessageSource(this);
		this.messageHandlers.add(handler);
	}
	
	@Override
	public boolean removeMessageHandler(MessageHandler handler) {
		boolean result = this.messageHandlers.remove(handler);
		if (result) handler.setMessageSource(null);
		return result;
	}
	
	@Override
	public Object notifyMessageHandlers(Message message) {
		Object result = getUnhandled();
		for (MessageHandler handler : this.messageHandlers) 
			if (handler.wantsToHandle(message)) {
				result = handler.handleMessage(message);
				break;
			}
		return result;
	}

	private void setRunning(boolean isRunning) {
		this.isRunning = isRunning;
		this.changed(Aspect.IS_RUNNING);
	}
	
	public static Object getUnhandled() {
		return "unhandled";
	}
}
