package fr.cnrs.mri.util;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.ArrayList;
import java.util.zip.CRC32;
import java.util.zip.Checksum;
import java.util.zip.Deflater;
import java.util.zip.ZipException;
import java.util.zip.ZipOutputStream;

public class ZIP32Util extends ZIPUtil {

	public static final long MAX_ZIP32_SIZE = 4000000000l;
	public static final int MAX_ZIP32_ENTRIES = 65535;
	
	/**
	 * the size of a chunk to calculate the checksum during
	 * zip creation, in bytes
	 */
	public static final int CHECKSUM_CHUNK_SIZE = 30 * 1000000;
	
	public ZIP32Util(OutputStream outputStream, ArrayList<ZipEntryWithPath> entries) {
		super();
		this.setOutputStream(outputStream);
		this.setZipEntries(entries);
	}
	
	public ZIP32Util(File outputFile, ArrayList<ZipEntryWithPath> entries) {
		super();
		System.out.println("Using Zip32Util");
		this.setOutputFile(outputFile);
		this.setZipEntries(entries);
		try {
			this.setOutputStream(new FileOutputStream(this.getOutputFile()));
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		}
	}

	private File getOutputFile() {
		return this.outputFile;
	}

	private void setOutputFile(File outputFile) {
		this.outputFile = outputFile;
		
	}

	public long zipFilesAndWriteToStream() throws IOException {
		ZipOutputStream zipOutput = new ZipOutputStream(this.getOutputStream());
		zipOutput.setLevel(this.getCompressionLevel());
		if (this.getCompressionLevel().equals(Deflater.NO_COMPRESSION)) {
			zipOutput.setMethod(ZipOutputStream.STORED);
		}
		long totalEntriesSize = 0;
		for (ZipEntryWithPath entry : this.getZipEntries()) {
			System.out.println("Zipping " + entry.getName() + " with path " + entry.getPath());
			
			// If no compression calculate the checksum for the entry
			if (this.getCompressionLevel().equals(Deflater.NO_COMPRESSION)) {
				Checksum checksum = new CRC32();
				byte[] chunk = new byte[CHECKSUM_CHUNK_SIZE];
				InputStream is = null;	
				int read = 0;
				is = new FileInputStream(new File(entry.getPath()));
				while ((read=is.read(chunk, 0, CHECKSUM_CHUNK_SIZE))>=0) {
					checksum.update(chunk, 0, read);
					if (outputFile == null) {
						totalEntriesSize += read;
						this.notifyProgress(totalEntriesSize);
					}
				}
				is.close();
				entry.setCrc(checksum.getValue());
				long entrySize = new File(entry.getPath()).length();
				entry.setCompressedSize(entrySize);
			}
			
			// trying to add the entry
			try {
				zipOutput.putNextEntry(entry);
			} catch (ZipException zipExcepion) {
				if (zipExcepion.getMessage().contains("duplicate")) continue;
			}
			if (entry.isDirectory()) {
				zipOutput.closeEntry();
				continue;
			}
			
			// Add the data to the zip, file by file
			FileInputStream fileInput = new FileInputStream(entry.getPath());
			byte[] data = new byte[bufferSize];
			int pos = 0;
			int i = 0;
			while ((pos = fileInput.read(data)) > 0) {
				zipOutput.write(data, 0, pos);
				totalEntriesSize += pos;
				if (i % 4 == 0) this.notifyProgress(totalEntriesSize);
				i++;
				if (isCanceled()) {
					zipOutput.closeEntry();
					fileInput.close();
					return 0;
				}
			}
			zipOutput.closeEntry();
			fileInput.close();
			this.notifyProgress(totalEntriesSize);
		}
		zipOutput.finish();
		return totalEntriesSize;
	}
	
	public boolean canHandleData() {
		return true;
//		long size = this.getTotalFileSize();
//		long items = this.getZipEntries().size();
//		boolean result = (size < MAX_ZIP32_SIZE) && (items < MAX_ZIP32_ENTRIES);
//		return result;
	}
	
	public OutputStream getOutputStream() {
		return outputStream;
	}

	private void setOutputStream(OutputStream outputStream) {
		this.outputStream = outputStream;
	}

	@Override
	public long compress() {
		try {
			return this.zipFilesAndWriteToStream();
		} catch (IOException e) {
			e.printStackTrace();
		}
		return 0;
	}
}
